from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import reduce
from pathlib import Path
from typing import Any, Mapping, Type, TypeVar

from confz import ConfZ, ConfZDataSource, ConfZFileSource, ConfZSource

from flexigurator.config_patch import patch_config

ConfigSource = ConfZSource | list[ConfZSource]


T = TypeVar("T")


def _flatten(a: T | list[T], b: T | list[T]) -> list[T]:
    joined = [a] + [b]

    def expand(agg: list[T], item: T | list[T]) -> list[T]:
        if isinstance(item, list):
            return agg + item
        return agg + [item]

    result = reduce(expand, joined, [])  # type: ignore
    return result  # type: ignore


def is_config_source(source: Any):
    return (
        isinstance(source, ConfZSource)
        or (isinstance(source, dict) and all(isinstance(key, str) for key in source))
        or (isinstance(source, list) and all(isinstance(item, ConfZSource) for item in source))
    )


class _VersionCollection(ABC):
    """Hold a collection of configuration versions."""

    @abstractmethod
    def _load_sources(self) -> None:
        """Lazily load the config version sources."""

    def get(self, version_name: str) -> ConfigSource:
        self._load_sources()

        version_name_split = version_name.split(".", maxsplit=1)
        source_name = version_name_split[0]

        source = getattr(self, source_name, None)

        if not source:
            raise AttributeError(f"Version source does not exist: {source_name}")

        if isinstance(source, _VersionCollection):
            if len(version_name_split) != 2:
                raise ValueError(f'"{source_name}" is a collection!')

            return source.get(version_name_split[1])

        if isinstance(source, dict):
            return ConfZDataSource(source)

        if not is_config_source(source):
            raise AttributeError(f"Not a version source: {source_name}")

        return source  # type: ignore


class DirectorySource(_VersionCollection):
    """Hold a collection of version sources located in a folder.

    Args:
        path (Path): Path to the folder containing the `.yaml` configuration files

    """

    _path: Path

    def __init__(self, path: Path):
        super().__init__()
        self._path = path

    def _load_sources(self) -> None:
        self._set_versions(self._load_versions(self._path))

    @staticmethod
    def _load_versions(path: Path) -> Mapping[str, ConfZFileSource | DirectorySource]:
        versions = dict[str, ConfZFileSource | DirectorySource]()
        for sub_path in path.glob("*"):
            if sub_path.suffix == ".yaml":
                name = sub_path.stem
                versions[name] = ConfZFileSource(sub_path)
            if sub_path.is_dir():
                name = sub_path.name
                versions[name] = DirectorySource(sub_path)
        return versions

    def _set_versions(self, configs: Mapping[str, ConfigSource | DirectorySource]) -> None:
        for name, version_source in configs.items():
            setattr(self, name, version_source)


class ConfigVersions(_VersionCollection):
    """Hold various versions of configurations that can be loaded in a context manager.

    One should implement this class in a config versions class (e.g. `Configs`). Fields pointing to
    `ConfZSource`s, `DirectorySource`s, or dictionaries can then be added to load configuration
    versions. These versions can then be loaded using `Configs.version` as a context manager.

    Example:
        class Config(ConfZ):  # type: ignore
            a: int
            b: int

        class Configs(ConfigVersions):
            CONFIG_CLASS = Config
            BASE = dict(a=1, b=2)  # Optional base version which is loaded before other versions
            test = ConfZFileSource("configs/test.yaml")

        with Configs().version("test"):
            print(Config())  # Configuration from the ConfZFileSource is now loaded

    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def _load_sources(self) -> None:
        pass

    @contextmanager
    def version(self, version_name: str, config_class: Type[ConfZ] | None = None):
        """Select a version from the collection and patch the supplied config class.

        Args:
            version_name (str): The name of the configuration version (i.e. the field name)
            config_class (Type[ConfZ]): The configuration class to patch

        Yields:
            ...

        Raises:
            AttributeError: When no config class is given or the version does not exist
        """
        config_class = config_class or getattr(self, "CONFIG_CLASS", None)

        if not config_class:
            raise AttributeError(
                'Need to supply "config_class" as parameter or "CONFIG_CLASS" in ConfigVersions!'
            )

        version_sources = self.get(version_name)

        if hasattr(self, "BASE"):
            version_base = self.get("BASE")
            version_sources = _flatten(version_base, version_sources)

        with patch_config(config_class, version_sources):
            yield



import json
import os
import warnings
from dataclasses import dataclass
from importlib import resources
from pathlib import Path
from typing import Any, Type

import mmh3
import yaml
from fastapi import FastAPI, Request, Response
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel

# Default location for Jinja templates is in the jinja_templates package
# This slightly convoluted method is used to get the path from the context manager
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    with resources.path("flexigurator.form", "jinja_templates") as _jinja_templates_file:
        _JINJA_TEMPLATE_DEFAULT_PATH = _jinja_templates_file


@dataclass
class ConfigTemplate:
    uid: str
    name: str
    path: Path

    @staticmethod
    def from_path(path: Path, templates_path: Path):
        # Create the name from the path by taking its path relative to the templates dir and
        # removing the suffix (e.g. '.yaml')
        # '/home/test/gridshield-python/configs/templates/sender/module.yaml' -> 'sender/module'
        name = os.path.splitext(path.relative_to(templates_path))[0]
        uid = str(mmh3.hash(name, signed=False))
        return ConfigTemplate(uid=uid, name=name, path=path)


def _load_config_templates(path: Path) -> list[ConfigTemplate]:
    paths = sorted(path.glob("**/*.yaml"), key=str)
    templates = [ConfigTemplate.from_path(file_path, path) for file_path in paths]
    templates = sorted(templates, key=lambda template: template.name)

    return templates


def _load_yaml(path: Path) -> dict[str, Any]:
    with open(path, "r", encoding="utf-8") as yaml_file:
        yaml_file_contents = yaml_file.read()

    yaml_object = yaml.safe_load(yaml_file_contents)

    return yaml_object if yaml_object else {}


def _config_form_start_vals(uid: str, config_templates: list[ConfigTemplate]) -> str:
    """Retrieves the start values for the template with the requested UID.

    Args:
        uid (str): The UID of the requested template
        config_templates (list[ConfigTemplate]): the list of templates

    Returns:
        str: a json string containing the start vals of the requested template
    """
    path = [template for template in config_templates if template.uid == uid][0].path
    return json.dumps(_load_yaml(path))


def _write_to_file(file_path: Path, contents: str) -> None:
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(contents)


def _save_config_form_output(json_: dict[str, Any], file_name: str, save_path: Path):
    """Writes the given json to a `.yaml` file.

    Args:
        json_ (dict[str, Any]): The json to be converted to yaml
        file_name (str): The file name without `.yaml` extention
        save_path (Path): The directory in which the file needs to be saved
    """
    if json_ == {}:
        yaml_str = ""
    else:
        yaml_str = yaml.dump(json_)

    yaml_file_path = save_path.joinpath(Path(file_name + ".yaml"))
    yaml_file_path.parent.mkdir(parents=True, exist_ok=True)
    _write_to_file(yaml_file_path, yaml_str)


def ConfigForm(
    config: Type[BaseModel],
    config_save_path: Path,
    config_templates_path: Path,
    jinja_templates_path: Path | None = None,
) -> FastAPI:  # pragma: no cover
    app = FastAPI()

    # Setup Jinja templates folder
    templates_dir_path = jinja_templates_path or _JINJA_TEMPLATE_DEFAULT_PATH
    templates = Jinja2Templates(directory=templates_dir_path)

    # Setup config templates and config json schema
    config_templates = _load_config_templates(config_templates_path)
    schema = config.schema_json()

    @app.get("/", response_model=None)
    async def root(request: Request) -> Response:
        # The landing page for the configurator.
        template_dict = {template.uid: template.name for template in config_templates}

        return templates.TemplateResponse(
            "index.html", {"request": request, "template_names": template_dict}
        )

    @app.get("/config_template/{uid}", response_model=None)
    async def config_form(request: Request, uid: str) -> Response:
        # Returns the form for the requested template.
        return templates.TemplateResponse(
            "config_form.html",
            {
                "request": request,
                "schema_json": schema,
                "start_val": _config_form_start_vals(uid, config_templates),
            },
        )

    @app.post("/config_json/{file_name}", response_model=None)
    async def _config_json(request: Request, file_name: str) -> dict[str, str]:
        # Writes the form results to disk.
        json_ = await request.json()
        _save_config_form_output(json_, file_name, config_save_path)
        return {"message": f"Parsed config json {json_}!"}

    return app



from flexigurator.config_patch import patch_config
from flexigurator.config_versions import ConfigVersions, DirectorySource
from flexigurator.form.form import ConfigForm
from flexigurator.placeholder import NotConfiguredError, placeholder



from contextlib import contextmanager
from typing import Any, Iterator, Type

from confz import ConfZ, ConfZDataSource, ConfZSource


@contextmanager
def patch_config(
    config_class: Type[ConfZ], data: dict[str, Any] | ConfZSource | list[ConfZSource]
) -> Iterator[None]:
    """Patch a config class with additional sources.

    This context manager temporarily updates the given config class with the provided data sources.
    This is useful for testing where a specific set of configuration values needs to be changed for
    a single test instance. Importantly, the provided configuration does not need to be complete,
    as long as the original configuration sources are.

    Args:
        config_class (Type[ConfZ]): The ConfZ config class
        data (dict[str, Any] | ConfZSource | list[ConfZSource]): A source of configuration in the
            form of a dictionary or one or more ConfZ sources

    Yields:
        None: This context manager does not yield

    """
    if isinstance(data, dict):
        data = ConfZDataSource(data)

    if not isinstance(data, list):
        data = [data]

    original_sources = config_class.CONFIG_SOURCES
    patched_sources: list[ConfZSource]

    if original_sources is None:
        patched_sources = []
    elif not isinstance(original_sources, list):
        patched_sources = [original_sources]
    else:
        patched_sources = list(original_sources)

    patched_sources += data

    with config_class.change_config_sources(patched_sources):
        yield



from typing import Any, Type

from pydantic import BaseModel
from pydantic.fields import Field, ModelField


class NotConfiguredError(Exception):
    """Raised when a requested configuration variable is not supplied in configuration source."""


class Placeholder(BaseModel):
    model_type: Type[BaseModel]
    model_fields: dict[str, ModelField] = Field(default_factory=dict)

    def __getattribute__(self, item: str) -> Any:
        # Get an attribute 'as usual' but raise a `NotConfiguredError` on Model fields.
        try:
            # Check if the requested attribute is part of the BaseModel fields
            fields = object.__getattribute__(self, "model_fields")

            if item in fields:
                # A BaseModel field is being requested, but as this model is not configured we
                # throw an exception
                raise NotConfiguredError(self.model_type)

        except AttributeError:
            # We cannot check using other means than a try/catch due to recursion.
            pass

        # The requested attribute is not a BaseModel field, so we return the attribute normally
        return object.__getattribute__(self, item)

    @classmethod
    def create(cls, model_type: Type[BaseModel]) -> "Placeholder":
        return Placeholder(model_type=model_type, model_fields=model_type.__fields__)

    def __repr__(self):
        return f"NotConfigured({self.model_type})"

    def dict(self, *args, **kwargs) -> Any:  # pylint: disable=W0613
        # Overrides the pydantic default `dict` to return nothing.
        return {}

    class Config:
        arbitrary_types_allowed = True


def placeholder(model_type: Type[BaseModel]) -> Any:
    """Return a placeholder set as pydantic field value to make it optionally configurable.

    Specifically this removes the need to use `None` to make fields optional, thereby making
    configuration classes much easier to work with as convoluted `None` checking is no longer
    needed.

    class SomeSubModel(BaseModel):
        some_int: int

    class SomeModel(BaseModel):
        sub_model: SomeSubModel = placeholder(SomeSubModel)

    SomeModel().sub_model.some_int                                    # Raises NotConfiguredError
    SomeModel(sub_model=SomeSubModel(some_int=5)).sub_model.some_int  # Returns 5

    Args:
        model_type (Type[BaseModel]): The type of BaseModel it makes configurable

    Returns:
        Any: A Placeholder object
    """
    return Placeholder.create(model_type=model_type)
